package cn.com.qmth.examcloud.core.basic.service.impl;

import cn.com.qmth.examcloud.api.commons.enums.DataRuleType;
import cn.com.qmth.examcloud.api.commons.security.enums.RoleMeta;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.DynamicEnum;
import cn.com.qmth.examcloud.commons.helpers.DynamicEnumManager;
import cn.com.qmth.examcloud.commons.helpers.poi.ExcelReader;
import cn.com.qmth.examcloud.commons.util.ByteUtil;
import cn.com.qmth.examcloud.commons.util.PathUtil;
import cn.com.qmth.examcloud.commons.util.SHA256;
import cn.com.qmth.examcloud.core.basic.base.constants.BasicConsts;
import cn.com.qmth.examcloud.core.basic.dao.*;
import cn.com.qmth.examcloud.core.basic.dao.entity.*;
import cn.com.qmth.examcloud.core.basic.dao.enums.OrgProperty;
import cn.com.qmth.examcloud.core.basic.service.OrgService;
import cn.com.qmth.examcloud.core.basic.service.UserDataRuleService;
import cn.com.qmth.examcloud.core.basic.service.bean.OrgInfo;
import cn.com.qmth.examcloud.core.basic.service.bean.datarule.UserDataRuleForm;
import cn.com.qmth.examcloud.core.basic.service.cache.OrgCache;
import cn.com.qmth.examcloud.core.basic.service.cache.OrgPropertyCache;
import cn.com.qmth.examcloud.core.basic.service.cache.RootOrgCache;
import cn.com.qmth.examcloud.task.api.DataSyncCloudService;
import cn.com.qmth.examcloud.task.api.request.SyncOrgReq;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;

import javax.persistence.criteria.Predicate;
import java.io.File;
import java.util.*;
import java.util.Map.Entry;
import java.util.stream.Collectors;

@Service
public class OrgServiceImpl implements OrgService {

    @Autowired
    RoleRepo roleRepo;

    @Autowired
    OrgRepo orgRepo;

    @Autowired
    OrgCache orgCache;

    @Autowired
    RootOrgCache rootOrgCache;

    @Autowired
    UserRepo userRepo;

    @Autowired
    UserRoleRelationRepo userRoleRelationRepo;

    @Autowired
    OrgPropertyRepo orgPropertyRepo;

    @Autowired
    StudentRepo studentRepo;

    @Autowired
    DataSyncCloudService dataSyncCloudService;

    @Autowired
    OrgPropertyCache orgPropertyCache;

    @Autowired
    UserDataRuleService userDataRuleService;

    private static final String[] EXCEL_HEADER = new String[]{"机构名称", "机构代码", "联系人", "联系电话"};

    /*
     * 保存顶级机构
     *
     * @author WANGWEI
     *
     * @see
     * cn.com.qmth.examcloud.core.basic.service.OrgService#saveRootOrg(cn.com.
     * qmth.examcloud.core.basic.service.bean.OrgInfo)
     */
    @Override
    public OrgEntity saveRootOrg(OrgInfo orgInfo) {

        String code = orgInfo.getCode();
        String name = orgInfo.getName();
        Boolean enable = orgInfo.getEnable();
        String contacts = orgInfo.getContacts();
        String telephone = orgInfo.getTelephone();
        String domainName = orgInfo.getDomainName();
        String remark = orgInfo.getRemark();
        Map<String, String> properties = orgInfo.getProperties();

        if (StringUtils.isBlank(code)) {
            throw new StatusException("150001", "code is null");
        }
        if (StringUtils.isBlank(name)) {
            throw new StatusException("150002", "name is null");
        }

        if (StringUtils.isNotBlank(domainName)) {
            OrgEntity tempOrg = orgRepo.findByParentIdIsNullAndDomainName(domainName);
            if (null != tempOrg && !tempOrg.getCode().equals(code)) {
                throw new StatusException("150004", "域名被占用");
            }
        }

        if (null == enable) {
            enable = true;
        }

        long updateTime = 0L;
        OrgEntity orgEntity = orgRepo.findByParentIdIsNullAndCode(code);
        if (null == orgEntity) {
            OrgEntity tmp = new OrgEntity();
            tmp.setRootId(-1L);
            tmp.setCode(code);
            tmp.setEnable(enable);
            tmp.setName(name);
            orgEntity = orgRepo.save(tmp);
        } else {
            if (!orgEntity.getId().equals(orgEntity.getRootId())) {
                throw new StatusException("150003", "数据错误");
            }

            if (null != orgEntity.getUpdateTime()) {
                updateTime = orgEntity.getUpdateTime().getTime();
            }
        }

        orgEntity.setRootId(orgEntity.getId());
        orgEntity.setName(name);
        orgEntity.setEnable(enable);
        orgEntity.setContacts(contacts);
        orgEntity.setTelephone(telephone);
        orgEntity.setDomainName(domainName);
        orgEntity.setRemark(remark);

        OrgEntity saved = orgRepo.saveAndFlush(orgEntity);

        // 同步操作
        if (updateTime != saved.getUpdateTime().getTime()) {
            SyncOrgReq req = new SyncOrgReq();
            req.setEnable(saved.getEnable());
            req.setId(saved.getId());
            req.setName(saved.getName());
            req.setParentId(saved.getParentId());
            req.setRootId(saved.getRootId());
            req.setSyncType("update");
            dataSyncCloudService.syncOrg(req);
        }

        saveOrgProperties(saved.getId(), properties);

        orgCache.remove(saved.getId());
        if (null != saved.getDomainName()) {
            rootOrgCache.remove(saved.getDomainName());
        }

        return saved;
    }

    /*
     * 保存子机构
     *
     * @author WANGWEI
     *
     * @see
     * cn.com.qmth.examcloud.core.basic.service.OrgService#saveSubOrg(cn.com.
     * qmth.examcloud.core.basic.service.bean.OrgInfo)
     */
    @Override
    public OrgEntity saveSubOrg(OrgInfo orgInfo) {

        String code = orgInfo.getCode();
        String name = orgInfo.getName();
        Boolean enable = orgInfo.getEnable();
        String contacts = orgInfo.getContacts();
        String telephone = orgInfo.getTelephone();
        Long rootId = orgInfo.getRootId();
        Long parentId = orgInfo.getParentId();
        String remark = orgInfo.getRemark();
        Map<String, String> properties = orgInfo.getProperties();

        if (null == enable) {
            enable = true;
        }

        if (null == rootId) {
            throw new StatusException("150000", "rootId is null");
        }
        if (StringUtils.isBlank(code)) {
            throw new StatusException("150001", "code is null");
        }
        if (StringUtils.isBlank(name)) {
            throw new StatusException("150002", "name is null");
        }
        if (null == parentId) {
            throw new StatusException("150003", "parentId is null");
        }

        OrgEntity rootOrgEntity = GlobalHelper.getEntity(orgRepo, rootId, OrgEntity.class);
        if (null == rootOrgEntity) {
            throw new StatusException("150004", "rootId is wrong");
        }
        if (null != rootOrgEntity.getParentId()) {
            throw new StatusException("150005", "rootId is wrong");
        }

        OrgEntity parentOrgEntity = GlobalHelper.getEntity(orgRepo, parentId, OrgEntity.class);
        if (null == parentOrgEntity) {
            throw new StatusException("150006", "parentId is wrong");
        }

        if (!rootOrgEntity.getRootId().equals(parentOrgEntity.getRootId())) {
            throw new StatusException("150007", "parentId, rootId is wrong");
        }

        long updateTime = 0L;
        OrgEntity orgEntity = orgRepo.findByRootIdAndCode(rootId, code);
        if (null == orgEntity) {
            orgEntity = new OrgEntity();
            orgEntity.setRootId(rootId);
            orgEntity.setParentId(parentId);
            orgEntity.setCode(code);
        } else {
            if (null != orgEntity.getUpdateTime()) {
                updateTime = orgEntity.getUpdateTime().getTime();
            }
        }

        orgEntity.setName(name);
        orgEntity.setEnable(enable);
        orgEntity.setContacts(contacts);
        orgEntity.setTelephone(telephone);
        orgEntity.setRemark(remark);

        OrgEntity saved = orgRepo.saveAndFlush(orgEntity);

        // 同步操作
        if (updateTime != saved.getUpdateTime().getTime()) {
            SyncOrgReq req = new SyncOrgReq();
            req.setEnable(saved.getEnable());
            req.setId(saved.getId());
            req.setName(saved.getName());
            req.setParentId(saved.getParentId());
            req.setRootId(saved.getRootId());
            req.setSyncType("update");
            dataSyncCloudService.syncOrg(req);
        }

        saveOrgProperties(saved.getId(), properties);

        orgCache.remove(saved.getId());
        if (null != saved.getDomainName()) {
            rootOrgCache.remove(saved.getDomainName());
        }

        return saved;
    }

    @Override
    public void saveOrgProperties(Long orgId, Map<String, String> properties) {

        GlobalHelper.getPresentEntity(orgRepo, orgId, OrgEntity.class);

        Map<DynamicEnum, String> map = checkAndGetOrgProperties(properties);

        for (Entry<DynamicEnum, String> entry : map.entrySet()) {
            DynamicEnum de = entry.getKey();
            String value = entry.getValue();
            OrgPropertyEntity entity = orgPropertyRepo.findByOrgIdAndKeyId(orgId, de.getId());
            if (null == entity) {
                entity = new OrgPropertyEntity();
                entity.setOrgId(orgId);
                entity.setKeyId(de.getId());
            }
            if (StringUtils.isBlank(value)) {
                entity.setValue(null);
            } else {
                entity.setValue(value);
            }

            orgPropertyRepo.save(entity);
        }

        if (null != properties) {
            for (String key : properties.keySet()) {
                orgPropertyCache.remove(orgId, key);
            }
        }

    }

    /**
     * 方法注释
     *
     * @param properties
     * @return
     * @author WANGWEI
     */
    private Map<DynamicEnum, String> checkAndGetOrgProperties(Map<String, String> properties) {
        DynamicEnumManager manager = OrgProperty.getDynamicEnumManager();

        Map<DynamicEnum, String> map = Maps.newHashMap();
        if (null == properties) {
            return map;
        }
        for (Entry<String, String> entry : properties.entrySet()) {
            String key = entry.getKey();
            String value = entry.getValue();
            DynamicEnum de = null;
            try {
                de = manager.getByName(key);
            } catch (Exception e) {
                throw new StatusException("001004", "机构属性错误. key=" + key);
            }
            if (!de.isLegal(value)) {
                throw new StatusException("001200", "机构属性值非法. key=" + key);
            }

            if (null == value) {
                map.put(de, null);
            } else {
                map.put(de, value.trim());
            }
        }

        return map;
    }

    @Override
    public List<Map<String, Object>> importSubOrg(Long rootOrgId, Long userId, File file) {

        List<String[]> lineList = null;
        try {
            lineList = ExcelReader.readSheetBySax(PathUtil.getCanonicalPath(file), 1, 4);
        } catch (Exception e) {
            throw new StatusException("100110", "Excel 解析失败");
        }

        if (CollectionUtils.isEmpty(lineList)) {
            throw new StatusException("100111", "Excel无内容");
        }

        if (10001 < lineList.size()) {
            throw new StatusException("100112", "数据行数不能超过10000");
        }

        List<Map<String, Object>> failRecords = Collections
                .synchronizedList(new ArrayList<Map<String, Object>>());

        List<OrgEntity> orgList = Lists.newArrayList();

        for (int i = 0; i < lineList.size(); i++) {
            String[] line = lineList.get(i);
            if (0 == i) {
                if (headerError(line)) {
                    throw new StatusException("100111", "Excel表头错误");
                }
                continue;
            }

            boolean hasError = false;
            StringBuilder msg = new StringBuilder();

            OrgEntity org = new OrgEntity();
            org.setParentId(rootOrgId);
            org.setEnable(true);

            String name = trimAndNullIfBlank(line[0]);
            if (StringUtils.isBlank(name)) {
                msg.append("  学习中心名称不能为空");
                hasError = true;
            } else if (name.length() > 30) {
                msg.append("  学习中心名称不能超过30个字符");
                hasError = true;
            }
            org.setName(name);

            String code = trimAndNullIfBlank(line[1]);
            if (StringUtils.isBlank(code)) {
                msg.append("  学习中心代码不能为空");
                hasError = true;
            } else if (code.length() > 30) {
                msg.append("  学习中心代码不能超过30个字符");
                hasError = true;
            }
            org.setCode(code);

            String contacts = trimAndNullIfBlank(line[2]);
            if (StringUtils.isBlank(contacts)) {
                msg.append("  联系人不能为空");
                hasError = true;
            } else if (contacts.length() > 30) {
                msg.append("  联系人不能超过30个字符");
                hasError = true;
            }
            org.setContacts(contacts);

            String telephone = trimAndNullIfBlank(line[3]);
            if (StringUtils.isBlank(telephone)) {
                msg.append("  联系电话不能为空");
                hasError = true;
            } else if (telephone.length() > 30) {
                msg.append("  联系电话不能超过30个字符");
                hasError = true;
            }
            org.setTelephone(telephone);

            if (hasError) {
                failRecords.add(newError(i + 1, msg.toString()));
            } else {
                orgList.add(org);
            }

        }

        if (CollectionUtils.isNotEmpty(failRecords)) {
            return failRecords;
        }

        List<Long> orgIds = new ArrayList<>();
        for (OrgEntity cur : orgList) {
            OrgInfo info = new OrgInfo();
            info.setCode(cur.getCode());
            info.setParentId(rootOrgId);
            info.setRootId(rootOrgId);

            info.setName(cur.getName());
            info.setContacts(cur.getContacts());
            info.setTelephone(cur.getTelephone());

            OrgEntity saved = saveSubOrg(info);

            orgIds.add(saved.getId());

            createLearnerCenterUser(saved);
        }

        // 用户与学习中心的数据权限
        UserDataRuleForm form = new UserDataRuleForm();
        form.setUserId(userId);
        form.setType(DataRuleType.ORG);
        form.setRefIds(orgIds);
        userDataRuleService.addUserDataRules(form);

        return failRecords;
    }

    @Override
    public Map<Long, String> findRootOrgNamesByIds(Set<Long> rootOrgIds) {
        if (CollectionUtils.isEmpty(rootOrgIds)) {
            return new HashMap<>();
        }

        Specification<OrgEntity> spec = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();
            predicates.add(root.get("id").in(rootOrgIds));
            predicates.add(cb.isNull(root.get("parentId")));
            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        List<OrgEntity> entities = orgRepo.findAll(spec);
        return entities.stream().collect(Collectors.toMap(v -> v.getId(), v -> v.getName(), (oldValue, newValue) -> newValue));
    }

    private void createLearnerCenterUser(OrgEntity org) {
        UserEntity user = userRepo.findByRootOrgIdAndLoginName(org.getRootId(), org.getCode());
        if (null != user) {
            return;
        }

        user = new UserEntity();
        user.setRootOrgId(org.getRootId());
        user.setOrgId(org.getId());
        user.setLoginName(org.getCode());
        user.setName(org.getContacts());
        user.setPhoneNumber(org.getTelephone());
        user.setEnable(true);
        user.setPasswordWeak(true);
        user.setPassword(ByteUtil.toHexAscii(SHA256.encode(user.getLoginName() + BasicConsts.DEFAULT_PASSWORD)));
        UserEntity savedUser = userRepo.save(user);

        List<UserRoleRelationEntity> userRoles = Lists.newArrayList();
        RoleEntity role = roleRepo.findByCode(RoleMeta.LC_USER.name());

        UserRoleRelationEntity relation = new UserRoleRelationEntity(savedUser.getId(), role.getId());
        userRoles.add(relation);
        userRoleRelationRepo.saveAll(userRoles);

        if (savedUser.getOrgId() != null) {
            // 用户与学习中心的数据权限
            userDataRuleService.addUserDataRule(savedUser.getId(), DataRuleType.ORG, savedUser.getOrgId());
        }
    }

    private Map<String, Object> newError(int lineNum, String msg) {
        Map<String, Object> map = Maps.newHashMap();
        map.put("lineNum", lineNum);
        map.put("msg", msg);
        return map;
    }

    private String trimAndNullIfBlank(String s) {
        if (StringUtils.isBlank(s)) {
            return null;
        }
        return s.trim();
    }

    /**
     * 方法注释
     *
     * @param header
     * @author WANGWEI
     */
    private boolean headerError(String[] header) {
        for (int i = 0; i < EXCEL_HEADER.length; i++) {
            if (null == header[i]) {
                return true;
            }
            if (!EXCEL_HEADER[i].equals(header[i].trim())) {
                return true;
            }
        }
        return false;
    }

}
